from typing import Dict, List

from pydantic import root_validator

from .. import Annotation, ResponseObject


class AnnotationsResponse(ResponseObject):
    """
    Response representing standoff annotations over a single stream of data (e.g. information extraction results).

    Attributes
    ----------
    type (string, required): type of response
    warnings (List[StatusMessage], optional): messages describing any warnings on responses
    features (dict, optional): arbitrary json metadata about content
    annotations (Dict[str, List[Annotation]], optional): optional annotations on request
    """

    type: str = "annotations"
    features: dict = None
    annotations: Dict[str, List[Annotation]] = None

    @root_validator()
    def either_features_or_annotations(cls, values):
        """ "
        validator: ensures either the "features" or "annotations" fields are present
        """
        features, annotations = values.get("features"), values.get("annotations")
        if features is None and annotations is None:
            raise ValueError('A annotation response must have either "features" or "annotations" fields')
        return values

    def auto_content(self):
        if self.warnings is not None:
            return self.dict()
        if self.features is not None:
            return self.dict()
        if len(self.annotations) == 1:
            return self.dict()["annotations"][0]
        return self.dict()["annotations"]



import sys
from argparse import ArgumentParser

from loguru import logger

from . import BaseELGCommand
from .docker_create import DockerCreateCommand


class DockerCommand(BaseELGCommand):
    @staticmethod
    def register_subcommand(parser: ArgumentParser):
        commands_parser = parser.add_parser("docker")
        subparsers = commands_parser.add_subparsers(
            title="ELG Docker CLI",
            description="elg docker <command> [<args>]",
            help="For more information about a command, run: `elg docker <command> --help`",
        )
        DockerCreateCommand.register_subcommand(subparsers)



import urllib
from typing import List, Union

import requests

from .entity import Entity
from .utils import API_URL
from .utils import ISO639 as iso639
from .utils import get_domain
from .utils.errors import catch_requests_error, ensure_response_ok

ISO639 = iso639()


class Catalog:
    """
    Class to use the ELG search API. Browse the ELG catalogue using Python.

    Examples::

        from elg import Catalog

        # First you have to init a catalog object.
        catalog = Catalog()

        # Then you can use the search method to search for resources. This method returns a list of Entity which can be displayed individually.
        # For example, we can search for a Machine Translation service for English and French.
        results = catalog.search(
            resource = "Tool/Service", # "Corpus", "Lexical/Conceptual resource" or "Language description"
            function = "Machine Translation", # function should be pass only if resource is set to "Tool/Service"
            languages = ["en", "fr"], # string or list if multiple languages
            limit = 100,
        )
        print(f"Machine Translation service for English and French:\\n{results[0]}")

        # Another example can be a German NER corpora.
        results = catalog.search(
            resource = "Corpus", # "Corpus", "Lexical/Conceptual resource" or "Language description"
            languages = ["German"], # string or list if multiple languages
            search="ner",
            limit = 100,
        )
        print(f"German corpus for NER:\\n{results[0]}")

        # You can init a service from an Entity.

        # We can use the catalog to search a Named Entity Recognizer for French and init a Service with the returned Entity.
        results = catalog.search(
            resource = "Tool/Service",
            function = "Named Entity Recognition",
            languages = ["fr"],
            limit = 1,
        )
        entity = results[0]
        print(entity)

        from elg import Service

        lt = Service.from_entity(entity=entity)
        result = lt("Jean Dupond vit à Paris.")
        print(f"\\n{result}")

    """

    def __init__(self, domain: str = "live"):
        """
        Init the Catalog object.

        Args:
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".
        """
        self.domain = get_domain(domain=domain)

    @catch_requests_error
    def _get(self, path: str, queries: List[set] = [], json: bool = False):
        """
        Internal method to call the API
        """
        url = (
            API_URL.format(self.domain)
            + path
            + ("?" if len(queries) >= 1 else "")
            + "&".join([f"{query}={urllib.parse.quote_plus(str(value))}" for (query, value) in queries])
        )
        response = requests.get(url)
        ensure_response_ok(response)
        if json:
            return response.json()
        return response

    def _search(
        self,
        entity: str = None,
        search: str = None,
        resource: str = None,
        function: str = None,
        languages: Union[str, list] = None,
        license: str = None,
        page: int = 1,
    ):
        """Internal method to send one search request to the API.

        Args:
            entity (str, optional): type of the entity to search. Can be 'LanguageResource', 'Organization', or
                'Project'. Defaults to None.
            search (str, optional): terms to use for the search request. Defaults to None.
            resource (str, optional): type of the language resource. Only used when the entity is set to
                'LanguageResource'. Can be 'Tool/Service', 'Lexical/Conceptual resource', 'Corpus', or
                'Language description'. Defaults to None.
            function (str, optional): type of the function of the service. Only used when resource set to 'Tool/Service'.
                Defaults to None.
            languages (Union[str, list], optional): language filter for the search request. Can be a string or
                a list of string. If it is a list of strings, the results of the request will match will all the languages
                and not one among all. The full name or the ISO639 code of the language can be used. Defaults to None.
            license (str, optional): license filter. Defaults to None.
            page (int, optional): page number. Defaults to 1.

        Returns:
            List[elg.Entity]: list of the results.
        """
        path = "search"
        queries = []
        queries.append(("entity_type_term", entity))
        queries.append(("page", page))
        if resource:
            queries.append(("resource_type__term", resource))
        if function:
            queries.append(("function__term", function))
        if search:
            queries.append(("search", search))
        if languages:
            if isinstance(languages, str):
                languages = languages.split(",")
            for language in languages:
                queries.append(("language__term", ISO639.LanguageName(language)))
        if license:
            queries.append(("licence__term", license))
        response = self._get(path=path, queries=queries, json=True)
        return [Entity.from_search_result(result=result, domain=self.domain) for result in response["results"]]

    def search(
        self,
        entity: str = "LanguageResource",
        search: str = None,
        resource: str = None,
        function: str = None,
        languages: Union[str, list] = None,
        license: str = None,
        limit: int = 100,
    ):
        """Method to send a search request to the API.

        Args:
            entity (str, optional): type of the entity to search. Can be 'LanguageResource', 'Organization', or
                'Project'. Defaults to "LanguageResource".
            search (str, optional): terms to use for the search request. Defaults to None.
            resource (str, optional): type of the language resource. Only used when the entity is set to
                'LanguageResource'. Can be 'Tool/Service', 'Lexical/Conceptual resource', 'Corpus', or
                'Language description'. Defaults to None.
            function (str, optional): type of the function of the service. Only used when resource set to 'Tool/Service'.
                Defaults to None.
            languages (Union[str, list], optional): language filter for the search request. Can be a string or
                a list of string. If it is a list of strings, the results of the request will match will all the languages
                and not one among all. The full name or the ISO639 code of the language can be used. Defaults to None.
            license (str, optional): license filter. Defaults to None.
            limit (int, optional): limit number of results. Defaults to 100.

        Returns:
            List[elg.Entity]: list of the results.

        Examples::

            results = catalog.search(
                resource = "Tool/Service",
                function = "Machine Translation",
                languages = ["en", "fr"],
                limit = 100,
            )
            results = catalog.search(
                resource = "Corpus",
                languages = ["German"],
                search="ner",
                limit = 100,
            )
        """
        results = []
        finished = False
        page = 1
        while len(results) < limit and finished is False:
            try:
                r = self._search(
                    entity=entity,
                    search=search,
                    resource=resource,
                    function=function,
                    languages=languages,
                    license=license,
                    page=page,
                )
            except:
                finished = True
                continue
            results.extend(r)
            page += 1
        return results[:limit]

    def interactive_search(
        self,
        entity: str = "LanguageResource",
        search: str = None,
        resource: str = None,
        function: str = None,
        languages: Union[str, list] = None,
        license: str = None,
    ):
        """
        Method to search resources interactivly. Warn: not well coded and tested.
        """
        page = 1
        while True:
            results = self._search(
                entity=entity,
                search=search,
                resource=resource,
                function=function,
                languages=languages,
                license=license,
                page=page,
            )
            self._display_entities(results)
            choice = self._make_choice(results)
            if choice == "n":
                page += 1
                continue
            else:
                break

    def _make_choice(self, entities: List[Entity]):
        """
        Internal method to search interactivly.
        """
        ids = [e.id for e in entities]
        choice = input("Display current page (c), display next page (n), type id to get more precision, stop (s): ")
        if choice == "c":
            self._display_entities(entities)
            return self._make_choice(entities)
        elif choice == "n":
            return "n"
        elif choice.isdigit():
            choice = int(choice)
            if choice in ids:
                entity = [e for e in entities if e.id == choice][0]
                print(entity)
                return self._make_choice(entities)
            else:
                print(f"Id {choice} is not in the list. Please try again.")
                return self._make_choice(entities)
        elif choice == "s":
            return "s"
        else:
            print("The choice you made has not been understood. Please try again.")
            return self._make_choice(entities)

    def _display_entities(self, entities: List[Entity]):
        """
        Internal method to display an entity list
        """
        print("-----------------------------------------------------------------------")
        print("Id         Name                                                        ")
        print("-----------------------------------------------------------------------")
        for e in entities:
            print(f"{e.id}\t   {e.resource_name:.60s}")
        print("-----------------------------------------------------------------------")



__version__ = "0.4.9"

import importlib.util

_docker_available = importlib.util.find_spec("docker") is not None
_flask_available = importlib.util.find_spec("flask") is not None
_flask_json_available = importlib.util.find_spec("flask_json") is not None
_quart_available = importlib.util.find_spec("quart") is not None
_aiohttp_available = importlib.util.find_spec("aiohttp") is not None
_requests_toolbelt_available = importlib.util.find_spec("requests_toolbelt") is not None


from .authentication import Authentication
from .benchmark import Benchmark
from .catalog import Catalog
from .corpus import Corpus
from .entity import Entity
from .pipeline import Pipeline
from .provider import Provider
from .service import Service

if _docker_available and _flask_available and _flask_json_available and _requests_toolbelt_available:
    from .flask_service import FlaskService

if _docker_available and _quart_available and _aiohttp_available and _requests_toolbelt_available:
    from .quart_service import QuartService



from pathlib import Path
from typing import Dict, List

from pydantic import validator

from .. import Annotation, Request


class AudioRequest(Request):
    """
    Request representing a piece of audio - the actual audio data will be sent as a separate request.

    Subclass of Request

    Attributes
    ----------
    type (str, required): the type of request. must be "audio"
    params (dict, optional): vendor specific requirements
    format (str, required): the format of the audio request
    sampleRate (int, required): sample rate of audio
    features (dict, optional): arbitrary json metadata about content
    annotations (Dict[str, List[Annotation]], optional): optional annotations on request
    """

    type: str = "audio"
    content: bytes = None
    format: str = "LINEAR16"
    sample_rate: int = None
    features: Dict = None
    annotations: Dict[str, List[Annotation]] = None

    @validator("format")
    def format_must_be_specific(cls, v):
        """
        validator: ensures the format of the audio request is either "LINEAR16" or "MP3"
        """
        acceptable_formats = ["LINEAR16", "MP3"]
        v = v.upper()
        if v not in acceptable_formats:
            raise ValueError("The format given is not supported")
        return v

    @classmethod
    def from_file(
        cls,
        filename,
        format: str = None,
        sample_rate: int = None,
        features: Dict = None,
        annotations: Dict[str, List[Annotation]] = None,
    ):
        filename = Path(filename)
        if not filename.is_file():
            raise ValueError(f"{filename} musts be the path to a file.")
        with open(filename, "rb") as f:
            content = f.read()
        if format is None:
            format = "MP3" if filename.suffix == ".mp3" else "LINEAR16"
        return cls(
            content=content,
            format=format,
            sample_rate=sample_rate,
            features=features,
            annotations=annotations,
        )

    def __str__(self):
        return " - ".join([f"{k}: {v}" for k, v in self.dict().items() if v is not None and k != "content"])



ENTRYPOINT_FLASK = """\
#!/bin/sh\nexec /sbin/tini -- venv/bin/gunicorn --bind=0.0.0.0:8000 "--workers=$WORKERS" --worker-tmp-dir=/dev/shm "$@" {service_script}:app\n"""

ENTRYPOINT_QUART = """\
#!/bin/sh\nexec /sbin/tini -- venv/bin/hypercorn --bind=0.0.0.0:8000 "$@" {service_script}:app\n"""

DOCKERFILE = """\
FROM python:slim

# Install tini and create an unprivileged user
ADD https://github.com/krallin/tini/releases/download/v0.19.0/tini /sbin/tini
RUN addgroup --gid 1001 "elg" && adduser --disabled-password --gecos "ELG User,,," --home /elg --ingroup elg --uid 1001 elg && chmod +x /sbin/tini

# Copy in just the requirements file
COPY --chown=elg:elg requirements.txt /elg/

# Everything from here down runs as the unprivileged user account
USER elg:elg

WORKDIR /elg

# Create a Python virtual environment for the dependencies
RUN python -m venv venv 
RUN /elg/venv/bin/python -m pip install --upgrade pip
RUN venv/bin/pip --no-cache-dir install -r requirements.txt 

# Copy ini the entrypoint script and everything else our app needs
{required_folders}
COPY --chown=elg:elg docker-entrypoint.sh {required_files} /elg/

ENV WORKERS=1

RUN chmod +x ./docker-entrypoint.sh
ENTRYPOINT ["./docker-entrypoint.sh"]\
"""

COPY_FOLDER = "COPY --chown=elg:elg {folder_name} /elg/{folder_name}/"



from typing import Callable, List, Union

from .authentication import Authentication
from .entity import Entity
from .model import Request
from .service import Service


class Pipeline:
    """
    Class to execute multiple services on after the other. The output of the first service is used as the input of the next one
    and so one. It is a basic approach but allows users to create complex services.

    Examples::

        from elg import Pipeline

        # A pipeline is simply a list of services and therefore can be initialized with a list of service ids or a list of entities.
        german_sentiment_analysis_pipeline = Pipeline.from_ids([607, 510])

        # To map the output of a service to the next one, we use the ouput_funcs parameter to extract the needed information
        # from the service response. In this example, we need to extract only the translated text to use it as input of the
        # sentiment analyser service. The first item of the output_funcs list corresponds to the callable object that extracts
        # the translated text from the first service output, and the second item of the list corresponds to the function apply
        # to the result of the second service. Here it is set the auto which means that the automatic content extraction will be used.
        results = german_sentiment_analysis_pipeline(
            "Ich habe diesen Film geliebt. Die Schauspieler, das Drehbuch: alles von einem Meisterwerk.",
            output_funcs=[
                lambda x: x.texts[0].content,
                "auto"
            ]
        )

        # The returned results object contains the result of each service of the pipeline.
        print("Result of the translation service: ", results[0])
        print("\\nResult of the sentiment analyser service: ", results[1])

        # With the pipeline class we can create a lot of different tools. For example, let's create a pipeline that generate an
        # German summary from an English news article. To do that, we will use a summarization service (478) and an English to
        # German Machine Translation service (610).
        pipeline = Pipeline.from_ids([478, 610])
        TEXT_TO_SUMMARIZE = '''In his new book, Fulfillment, Alec MacGillis writes of an Amazon distribution center in Sparrows Point, Maryland that sits on land once occupied by a Bethlehem Steel plant. The story underscores how dramatically the U.S. economy has transformed in recent years. Instead of making things, many of our biggest companies now distribute things made elsewhere. We've moved from an economy of production to one of dispersion. The shift from factory to fulfillment work is core to the American story right now. For the American worker, a factory job like one at Bethlehem Steel was dangerous, but it paid $30 to $40 per hour, and many stuck with it for life. At an Amazon Fulfillment Center, pay starts at $15 per hour, algorithms monitor your performance, and many workers leave soon after joining. "There's 100% turnover in the warehouses," MacGillis told me this week. "100% \every single year." Some blame the move to fulfillment work entirely on Amazon, but it didn't happen in a vacuum. American politicians helped it along by signing trade deals like NAFTA and enthusiastically welcoming China into the World Trade Organization - and doing so without sufficient safeguards. American industry then suffered the consequences, and Amazon reaped the benefits. Listen to Alec MacGillis on this week's Big Technology Podcast on Apple, Spotify, or your app of choice. The U.S. embrace of globalization flooded the country's markets with inexpensive products, and plants stateside couldn't keep up. As American factories went under or moved overseas, there was glaring a need for a company that could get the new, affordable products to Americans' doorsteps. Amazon eagerly stepped in. And its timing couldn't have been better. In January 1994, NAFTA opened up trade between U.S., Canada, and Mexico. Seven months later, Jeff Bezos founded Amazon. In 2000, Amazon launched its third-party marketplace. One year later, China joined the WTO.For workers, going from $40 to $15 per hour meant moving from an annual salary of $80,000 to one closer to $30,000. And as U.S. industry struggled, Amazon's hired 1,400 workers per day. Many American workers are now removed from the production process - employed by middlemen taking a cut - so the drop in wage is natural. MacGillis tells the story of some workers who, on the same land, worked for Bethlehem Steel and then Amazon for one-third the wage. American workers have thus taken a hard look at our political system and found it wanting. In the aftermath of the trade deals, they've been left too often to decide between unemployment and jobs at fulfillment centers, ride-hailing services, and app-based food delivery. In 2016, many workers remembered Bill Clinton signing NAFTA and backing China's entry to the WTO, and voted for Donald Trump. In 2020, as Covid raged, enough moved to Joe Biden's camp that he won the election with an amalgamation of workers and college-educated liberals, what MacGillis calls the "Amazon Coalition." In some cases, Amazon workers drawn to the Democratic party for its alliance with labor were peeing in bottles as they delivered packages to urban-dwellers drawn to the party for its social values. The underlying tension flared up recently as workers at an Amazon fulfillment center in Bessemer, Alabama pushed to unionize. Bernie Sanders and Elizabeth Warren came out strongly in support of the union, fully aware of how vital fulfillment workers are to their future. Amazon then mocked the government for failing the working class. Its executive overseeing the fulfillment centers, Dave Clark, said "We actually deliver a progressive workplace," while hailing Amazon's $15 minimum wage compared to the federal government's $7.25. The union effort in Bessemer seems en route to defeat. But even in victory, it would not have replaced the fulfillment center work with factory work. The reality facing U.S. workers today is that low-paying fulfillment jobs are often their best option. And we're likely to see more political instability if we don't find some way to generate jobs that give workers a shot at the middle class. '''

        results = pipeline(
            TEXT_TO_SUMMARIZE,
            output_funcs=[
                lambda x: " ".join([data.features["prefLabel"] for data in x.annotations["Main sentence"]]),
                lambda x: x.texts[0].content,
            ]
        )
        print("German summary: ", results[-1])

    """

    def __init__(self, services: List[Service]):
        """
        Init a Pipeline with the services to execute

        Args:
            services (List[Service]): list of the services to execute. This first service of
                the list is the first service executed.
        """
        for service in services:
            assert isinstance(service, Service)
        self.services = services

    @classmethod
    def from_entities(
        cls,
        entities: List[Entity],
        auth_object: str = None,
        auth_file: str = None,
        scope: str = None,
        use_cache: bool = True,
        cache_dir="~/.cache/elg",
    ):
        """
        Class method to init a Pipeline using a list of entities which will be convert into services
        using the `from_entity` class method of the Service class. Refer to this method for further explanation.
        """
        return cls(
            [
                Service.from_entity(
                    entity=entity,
                    auth_object=auth_object,
                    auth_file=auth_file,
                    scope=scope,
                    use_cache=use_cache,
                    cache_dir=cache_dir,
                )
                for entity in entities
            ]
        )

    @classmethod
    def from_ids(
        cls,
        ids: List[int],
        auth_object: Authentication = None,
        auth_file: str = None,
        scope: str = None,
        domain: str = None,
        use_cache: bool = True,
        cache_dir: str = "~/.cache/elg",
    ):
        """
        Class method to init a Pipeline using a list of ids which will be convert into services
        using the `from_id` class method of the Service class. Refer to this method for further explanation.
        """
        return cls(
            [
                Service.from_id(
                    id=id,
                    auth_object=auth_object,
                    auth_file=auth_file,
                    scope=scope,
                    domain=domain,
                    use_cache=use_cache,
                    cache_dir=cache_dir,
                )
                for id in ids
            ]
        )

    def __call__(
        self,
        request_input: Union[str, List[str], Request] = None,
        request_types: Union[str, List[str]] = "text",
        sync_mode: bool = False,
        timeout: int = None,
        check_file: bool = True,
        verbose: bool = True,
        output_funcs: Union[str, Callable, List[Union[str, Callable]]] = "auto",
    ):
        """
        Method to run the pipeline.

        Args:
            request_input (Union[str, List[str], Request], optional): input that corresponds to the request_input parameter of
                the Service __call__ method. Defaults to None.
            request_type (str, optional): precise the type of the request. Can be "text", "structuredText", or "audio".
                It is only used if request_input is not a Request object. Defaults to "text".
            sync_mode (bool, optional): sync_mode parameter to give to the Service __call__ method. Defaults to False.
            timeout (int, optional): timeout parameter to give to the Service __call__ method. Defaults to None.
            check_file (bool, optional): check_file parameter to give to the Service __call__ method. Defaults to True.
            output_funcs (Union[str, Callable, List[Union[str, Callable]]], optional): output_func parameters to give to the Services
                __call__ method. Defaults to "auto".

        Returns:
            Union[Dict, str]: service response in JSON format or as a string if output_func returns a string.
        """
        if isinstance(request_types, str):
            request_types = [request_types for _ in range(len(self.services))]
        if len(request_types) != len(self.services):
            raise ValueError("the lenght of request_type musts be equal to the number of services.")

        if isinstance(output_funcs, str) and output_funcs != "auto":
            raise ValueError(
                "output_funcs must be set to 'auto', a callable object, or a list of 'auto' or callable objects with the same length as the number of services in self.services."
            )
        elif isinstance(output_funcs, str) or isinstance(output_funcs, Callable):
            output_funcs = [output_funcs for _ in range(len(self.services))]
        elif isinstance(output_funcs, List):
            if len(output_funcs) != len(self.services):
                raise ValueError(
                    "output_funcs must be set to 'auto', a callable object, or a list of 'auto' or callable objects with the same length as the number of services in self.services."
                )
            for output_func in output_funcs:
                if not ((isinstance(output_func, str) and output_func == "auto") or isinstance(output_func, Callable)):
                    raise ValueError(
                        "output_funcs must be set to 'auto', a callable object, or a list of 'auto' or callable objects with the same length as the number of services in self.services."
                    )
        else:
            raise ValueError(
                "output_funcs must be set to 'auto', a callable object, or a list of 'auto' or callable objects with the same length as the number of services in self.services."
            )

        result = request_input
        results = []
        for service, request_type, output_func in zip(self.services, request_types, output_funcs):
            result = service(
                request_input=result,
                request_type=request_type,
                sync_mode=sync_mode,
                timeout=timeout,
                check_file=check_file,
                verbose=verbose,
                output_func=output_func,
            )
            results.append(result)
            if verbose:
                print(f"Result:\n\t{result}\n")
        return results



from pydantic import BaseModel


class Request(BaseModel):
    """
    Representation of a service invocation request

    Intended to be abstract, subclasses should be initiated with their specific type

    Subclasses
    ----------
    request.AudioRequest
    request.TextRequest
    request.StructuredTextRequest

    Attributes
    ----------
    type (str, required on subclass instantiation): the type of request
    params (dict, optional): vendor specific requirements
    """

    type: str = None
    params: dict = None

    def __str__(self):
        return " - ".join([f"{k}: {v}" for k, v in self.dict().items() if v is not None])



import random
from pathlib import Path

import requests

from .authentication import (Authentication, NeedAuthentication,
                             need_authentication)
from .utils import S3_UPLOAD_URL, XML_UPLOAD_URL, get_domain
from .utils.errors import ensure_response_ok


class Provider(NeedAuthentication):
    """
    Class to provide resources on ELG
    """

    def __init__(
        self,
        auth_object: Authentication = None,
        auth_file: str = None,
        scope: str = "openid",
        domain: str = "live",
        use_cache: bool = True,
        cache_dir: str = "~/.cache/elg",
    ):
        """
        Initialization method to create a Provider object. Authenticate the provider using the ELG login.

        Args:
            auth_object (elg.Authentication, optional): elg.Authentication object to use. Defaults to None.
            auth_file (str, optional): json file that contains the authentication tokens. Defaults to None.
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens. Defaults to "openid".
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".
            use_cache (bool, optional): True if you want to use cached files. Defaults to True.
            cache_dir (str, optional): path to the cache_dir. Set it to None to not store any cached files. Defaults to "~/.cache/elg".
        """
        self.domain = get_domain(domain)
        self._authenticate(
            auth_object=auth_object, auth_file=auth_file, scope=scope, use_cache=use_cache, cache_dir=cache_dir
        )

    @need_authentication()
    def upload_xml(
        self, xml_file: str, fonctional_service: bool = False, under_construction: bool = False, zip_file: str = None
    ):
        """
        Method to upload a metadata record.

        Args:
            xml_file (str): name of the XML file containing the information of the record to upload.
            fonctional_service (bool, optional): True if the record is a fonctional service. Defaults to False.
            under_construction (bool, optional): True if the record is under construction. Defaults to False.
            zip_file (str, optional): name of the data zip file to upload if the record is a corpus. Defaults to None.
        """
        headers = {
            "Authorization": f"Bearer {self.authentication.access_token}",
            "FUNCTIONAL-SERVICE": str(fonctional_service).lower(),
            "UNDER-CONSTRUCTION": str(under_construction).lower(),
        }
        with open(xml_file, "rb") as data:
            files = {"file": (Path(xml_file).name, data, "text/xml")}
            response = requests.post(XML_UPLOAD_URL.format(self.domain), files=files, headers=headers)
        ensure_response_ok(response=response)
        if isinstance(zip_file, str):
            self.upload_zip(pk=response.json().get("pk"), zip_file=zip_file)
        return response

    @need_authentication()
    def upload_zip(self, pk: int, zip_file: str):
        """
        Method to upload a data zip file link to a record
        """
        headers = {
            "Authorization": f"Bearer {self.authentication.access_token}",
            "ELG-RESOURCE-ID": str(pk),
            "filename": "und",
            "filetype": "und",
            "scope": "private",
        }
        with open(zip_file, "rb") as data:
            response = requests.put(S3_UPLOAD_URL.format(self.domain), data=data, headers=headers)
        ensure_response_ok(response=response)



import hashlib
import json
import pickle
import time
from functools import wraps
from os.path import expanduser
from pathlib import Path

import requests

from .utils import get_domain
from .utils.errors import (AuthenticationException,
                           RefreshTokenExpirationException,
                           catch_requests_error, ensure_response_ok)

ONE_HUNDRED_YEARS_IN_SECONDS = 100 * 365 * 24 * 3600


class Authentication:
    """
    Class to authenticate in the ELG using out-of-band authentication.

    Args:
        base_url (str): url to authenticate.
        token_url (str): url to request tokens.
        client (str): name of the Keycloak client.
        redirect_uri (str): redirect uri.

    """

    base_url = "{}/auth/realms/ELG/protocol/openid-connect/auth"
    token_url = "{}/auth/realms/ELG/protocol/openid-connect/token"
    client = "elg-oob"
    redirect_uri = "urn:ietf:wg:oauth:2.0:oob"

    def __init__(self, domain: str):
        """
        Init an Authentication object.

        Args:
            domain (str): ELG domain you want to use. "live" to use the public ELG, "dev" to use the development ELG
                and another value to use a local ELG.
        """
        self.access_token = None
        self.expires_in = None
        self.expires_time = time.gmtime(0)
        self.refresh_expires_in = None
        self.refresh_expires_time = time.gmtime(0)
        self.refresh_token = None
        self.token_type = None
        self.id_token = None
        self.not_before_policy = None
        self.session_state = None
        self.scope = None
        self.domain = get_domain(domain=domain)

    def __str__(self):
        return f"Authentication(valid_until: {time.strftime('%Y-%m-%dT%H:%M:%SZ', self.expires_time)})"

    def __repr__(self):
        obj = self.__dict__.copy()
        obj["access_token"] = "*****"
        obj["refresh_token"] = "*****"
        obj["id_token"] = "*****"
        return str(obj)

    @classmethod
    def init(cls, scope: str = "openid", domain: str = "live"):
        """
        Class method to init an Authentication object and authenticate to the ELG.

        Args:
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens. Defaults to "openid".
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".

        Returns:
            elg.Authentication: Authentication object with Keycloak authentication tokens.
        """
        authentication = cls(domain=domain)
        authentication.create(scope)
        return authentication

    @classmethod
    def from_json(cls, filename: str):
        """
        Class method to init an Authentication object from a json file.

        Args:
            filename (str): name of the json file.

        Returns:
            elg.Authentication: Authentication object with Keycloak authentication tokens.
        """
        with open(filename, "r") as tokens_file:
            tokens = json.load(tokens_file)
        authentication = cls(tokens["domain"])
        authentication.access_token = tokens["access_token"]
        authentication.expires_in = tokens["expires_in"]
        authentication.expires_time = time.struct_time(tokens["expires_time"])
        authentication.refresh_expires_in = tokens["refresh_expires_in"]
        authentication.refresh_expires_time = time.struct_time(tokens["refresh_expires_time"])
        authentication.refresh_token = tokens["refresh_token"]
        authentication.token_type = tokens["token_type"]
        authentication.id_token = tokens["id_token"]
        authentication.not_before_policy = tokens["not_before_policy"]
        authentication.session_state = tokens["session_state"]
        authentication.scope = tokens["scope"]
        return authentication

    @classmethod
    def create_authentication_url(cls, scope: str = "openid", domain: str = "live") -> str:
        """
        Class method to create the Keycloak authentication url.

        Args:
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens. Defaults to "openid".
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".

        Returns:
            str: Keycloak authentication url.
        """
        domain = get_domain(domain=domain)
        queries = [
            ("client_id", cls.client),
            ("redirect_uri", cls.redirect_uri),
            ("response_type", "code"),
            ("scope", scope),
        ]
        auth_url = (
            cls.base_url.format(domain)
            + ("?" if len(queries) >= 1 else "")
            + "&".join([f"{query}={str(value)}" for (query, value) in queries])
        )
        return auth_url

    @classmethod
    def from_success_code(cls, code: str, domain: str = "live"):
        """
        Class method to init an Authentication object from a success code.

        Args:
            code (str): success code obtained after authentication.
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".

        Returns:
            elg.Authentication: Authentication object with Keycloak authentication tokens.
        """
        authentication = cls(domain=domain)
        data = {
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": cls.redirect_uri,
            "client_id": cls.client,
        }
        authentication._requesting_oauth_token(data)
        return authentication

    def create(self, scope: str = "openid"):
        """
        Method to create Keycloak authentication tokens.

        Args:
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens. Defaults to "openid".
        """
        queries = [
            ("client_id", self.client),
            ("redirect_uri", self.redirect_uri),
            ("response_type", "code"),
            ("scope", scope),
        ]
        auth_url = (
            self.base_url.format(self.domain)
            + ("?" if len(queries) >= 1 else "")
            + "&".join([f"{query}={str(value)}" for (query, value) in queries])
        )

        print("Please go to this URL in your browser: {}\n".format(auth_url))
        auth_code = input('Paste the "success code": ')
        data = {
            "grant_type": "authorization_code",
            "code": auth_code,
            "redirect_uri": self.redirect_uri,
            "client_id": self.client,
        }
        self._requesting_oauth_token(data=data)

    def refresh(self):
        """
        Method to refresh to access_token using the refresh_token.

        Raises:
            AuthenticationException: There is no refresh_token to refresh the access_token.
        """
        if self.refresh_token is None:
            raise AuthenticationException()
        data = {"grant_type": "refresh_token", "client_id": self.client, "refresh_token": self.refresh_token}
        self._requesting_oauth_token(data=data)

    def refresh_if_needed(self):
        """
        Method that call the refresh method only if needed, i.e. the access_token is expired.

        Raises:
            RefreshTokenExpirationException: The refresh_token is expired.
        """
        # refresh if the token will expire in less than 60 seconds
        # idea from: https://stackoverflow.com/questions/30826726/how-to-identify-if-the-oauth-token-has-expired
        if time.gmtime(time.time() + 60) > self.expires_time:
            if time.gmtime() > self.refresh_expires_time:
                raise RefreshTokenExpirationException()
            if time.gmtime(time.time() + 3600) > self.refresh_expires_time:
                token_expire_time = time.mktime(time.gmtime()) - time.mktime(self.refresh_expires_time)
                print(f"Warning: The refresh token will expire in {token_expire_time} seconds!")
            self.refresh()

    @catch_requests_error
    def _requesting_oauth_token(self, data):
        """
        Method to obtain the Keycloak authentication tokens from data.

        Args:
            data (dict): data used to ask the tokens to Keycloak.
        """
        response = requests.post(self.token_url.format(self.domain), data=data)
        ensure_response_ok(response=response)
        tokens = response.json()
        self.access_token = tokens["access_token"]
        self.expires_in = tokens["expires_in"]
        self.expires_time = time.gmtime(time.time() + self.expires_in)
        self.refresh_expires_in = tokens["refresh_expires_in"]
        self.refresh_expires_time = time.gmtime(
            time.time() + (self.refresh_expires_in if self.refresh_expires_in != 0 else ONE_HUNDRED_YEARS_IN_SECONDS)
        )
        self.refresh_token = tokens["refresh_token"]
        self.token_type = tokens["token_type"]
        self.id_token = tokens.get(
            "id_token"
        )  # use get to not raise an error if there is no id_token (for offline_access)
        self.not_before_policy = tokens["not-before-policy"]
        self.session_state = tokens["session_state"]
        self.scope = tokens["scope"]

    def to_json(self, filename: str):
        """
        Save the Keycloak authentication tokens to a json file.

        Args:
            filename (str): name of the json file.
        """
        with open(filename, "w") as tokens_file:
            json.dump(self.__dict__, tokens_file, indent="\t")


class NeedAuthentication:
    """
    Parent class for class which needs authentication. Provide useful methods
    """

    def _authenticate(self, auth_object: Authentication, auth_file: str, scope: str, use_cache: str, cache_dir: str):
        """
        Method to authenticate.

        Args:
            auth_object (elg.Authentication, optional): elg.Authentication object to use.
            auth_file (str, optional): json file that contains the authentication tokens.
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens.
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG.
            use_cache (bool, optional): True if you want to use cached files.
            cache_dir (str, optional): path to the cache_dir. Set it to None to not store any cached files.

        Raises:
            DomainException: the domain of the authentication is different from the service domain.
        """

        # Use the authentification object if given
        if auth_object is not None:
            print(f"Using authentication object")
            assert isinstance(
                auth_object, Authentication
            ), f"auth_object ({type(auth_object)}) must be an Authentication object."
            if auth_file is not None:
                print("Warning: auth_file argument will not be used as an auth_object has been passed.")
            if scope is not None:
                print(
                    "Warning: scope argument saved in the authentication object will overwrite the "
                    "scope argument you put."
                )
            self.authentication = auth_object
            if self.domain != self.authentication.domain:
                raise DomainException(self.domain, self.authentication.domain)
            if not self._is_auth_object_valid(self.authentication):
                raise RefreshTokenExpirationException()

        # Use the authentication file if given
        elif isinstance(auth_file, str):
            print(f"Using authentication file: {auth_file}")
            if scope is not None:
                print(
                    "Warning: scope argument saved in the authentication filename will overwrite the "
                    "scope argument you put."
                )
            self.authentication = Authentication.from_json(filename=auth_file)
            if self.domain != self.authentication.domain:
                raise DomainException(self.domain, self.authentication.domain)
            if not self._is_auth_object_valid(self.authentication):
                raise RefreshTokenExpirationException()

        # Use a cached authentification object or create a new one
        else:
            scope = scope if scope is not None else "openid"  # default scope

            if cache_dir is not None and use_cache is True:
                cache_path = expanduser(cache_dir)
                cache_path = Path(cache_path) / Path("tokens")
                cache_path.mkdir(parents=True, exist_ok=True)
                authentication_hash = hashlib.sha256(
                    bytes(f"authentication - {scope} - {self.domain}", "utf-8")
                ).hexdigest()
                cache_filename = cache_path / Path(authentication_hash)
                if cache_filename.is_file():
                    with open(cache_filename, "rb") as cache_file:
                        self.authentication = pickle.load(cache_file)

            if (not hasattr(self, "authentication")) or (not self._is_auth_object_valid(self.authentication)):
                self.authentication = Authentication.init(scope=scope, domain=self.domain)

        if cache_dir is not None:
            cache_path = expanduser(cache_dir)
            cache_path = Path(cache_path) / Path("tokens")
            cache_path.mkdir(parents=True, exist_ok=True)
            authentication_hash = hashlib.sha256(
                bytes(
                    f"authentication - {'openid' if 'openid' in self.authentication.scope else 'offline_access'} - {self.authentication.domain}",
                    "utf-8",
                )
            ).hexdigest()
            cache_filename = cache_path / Path(authentication_hash)
            with open(cache_filename, "wb") as cache_file:
                pickle.dump(self.authentication, cache_file)

    def _is_auth_file_valid(self, auth_file: str, scope: str = None) -> bool:
        """
        Method to check is an authentication file contains valid authentication information.

        Args:
            auth_file (str): json file that contains the authentication tokens.
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens.

        Returns:
            bool: True is the authentication file contains valid information, else False.
        """
        refresh_expires_time, auth_domain, auth_scope = get_argument_from_json(
            filename=auth_file, args=["refresh_expires_time", "domain", "scope"]
        )
        # verify if tokens are valid
        if time.gmtime() > time.struct_time(refresh_expires_time):
            return False
        # verify if the domain of cached tokens corresponds to the desire domain
        if self.domain != auth_domain:
            return False
        # verify if the scope of cached tokens corresponds to the desire scope
        if scope is not None and scope not in auth_scope:
            return False
        return True

    def _is_auth_object_valid(self, auth_object: Authentication, scope: str = None) -> bool:
        """
        Method to check is an authentication object contains valid authentication information.

        Args:
            auth_object (Authentication): authentication object.
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens.

        Returns:
            bool: True is the authentication object contains valid information, else False.
        """
        # verify if tokens are valid
        if time.gmtime() > time.struct_time(auth_object.refresh_expires_time):
            return False
        # verify if the domain of cached tokens corresponds to the desire domain
        if self.domain != auth_object.domain:
            return False
        # verify if the scope of cached tokens corresponds to the desire scope
        if scope is not None and scope not in auth_object.scope:
            return False
        return True


def need_authentication():
    """
    Decorator for methods to refresh to authentication tokens before calling the method
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            self.authentication.refresh_if_needed()
            result = func(self, *args, **kwargs)
            return result

        return wrapper

    return decorator
